from x_transformers import TransformerWrapper, Decoder
from x_transformers.autoregressive_wrapper import AutoregressiveWrapper

import random
import tqdm
import gzip
import numpy as np
import torch
import torch.optim as optim
from torch.nn import functional as F
from torch.utils.data import DataLoader, Dataset

# constants

NUM_BATCHES = int(1e5)
BATCH_SIZE = 4
GRADIENT_ACCUMULATE_EVERY = 4
LEARNING_RATE = 1e-4
GENERATE_EVERY  = 500
GENERATE_LENGTH = 256
SEQ_LEN = 256

VALIDATE_EVERY  = 100
VALIDATE_SEQ_LENS = (256, 512, 1024, 2048, 4096)

# helpers

def cycle(loader):
    while True:
        for data in loader:
            yield data

def decode_token(token):
    return str(chr(max(32, token)))

def decode_tokens(tokens):
    return ''.join(list(map(decode_token, tokens)))

# instantiate GPT-like decoder model

model = TransformerWrapper(
    num_tokens = 256,
    max_seq_len = SEQ_LEN,
    use_abs_pos_emb = False,
    attn_layers = Decoder(
        dim = 512,
        depth = 6,
        heads = 8,
        dynamic_pos_bias = True,
    )
)

model = AutoregressiveWrapper(model)
model.cuda()

# prepare enwik8 data

with gzip.open('./data/enwik8.gz') as file:
    data = np.frombuffer(file.read(int(95e6)), dtype=np.uint8).copy()
    train_x, valid_x = np.split(data, [int(90e6)])
    data_train, data_val = torch.from_numpy(train_x), torch.from_numpy(valid_x)

class TextSamplerDataset(Dataset):
    def __init__(self, data, seq_len):
        super().__init__()
        self.data = data
        self.seq_len = seq_len

    def __getitem__(self, index):
        rand_start = torch.randint(0, self.data.size(0) - self.seq_len - 1, (1,))
        full_seq = self.data[rand_start: rand_start + self.seq_len + 1].long()
        return full_seq.cuda()

    def __len__(self):
        return self.data.size(0) // self.seq_len

train_dataset = TextSamplerDataset(data_train, SEQ_LEN)
train_loader  = cycle(DataLoader(train_dataset, batch_size = BATCH_SIZE, drop_last = True))

val_dataset_generate = TextSamplerDataset(data_val, SEQ_LEN)

# validation loaders with different sequence lengths

val_loaders = dict()

for valid_seq_len in VALIDATE_SEQ_LENS:
    val_dataset   = TextSamplerDataset(data_val, valid_seq_len)
    val_loader    = cycle(DataLoader(val_dataset, batch_size = BATCH_SIZE, drop_last = True))

    val_loaders[valid_seq_len] = val_loader

# optimizer

optim = torch.optim.Adam(model.parameters(), lr=LEARNING_RATE)

# training

for i in tqdm.tqdm(range(NUM_BATCHES), mininterval=10., desc='training'):
    model.train()

    for __ in range(GRADIENT_ACCUMULATE_EVERY):
        loss = model(next(train_loader))
        (loss / GRADIENT_ACCUMULATE_EVERY).backward()

    print(f'training loss: {loss.item()}')

    torch.nn.utils.clip_grad_norm_(model.parameters(), 0.5)
    optim.step()
    optim.zero_grad()

    if i % VALIDATE_EVERY == 0:
        print(f'validation losses:\n')

        model.eval()
        with torch.no_grad():
            for valid_seq_len in VALIDATE_SEQ_LENS:
                val_loader = val_loaders[valid_seq_len]

                loss = model(next(val_loader))
                print(f'[{valid_seq_len}]:\t {loss.item()}')

        print('\n')

    if i % GENERATE_EVERY == 0:
        model.eval()
        inp = random.choice(val_dataset_generate)[:-1]
        prime = decode_tokens(inp)
        print(f'%s \n\n %s', (prime, '*' * 100))

        sample = model.generate(
            prompts = inp,
            seq_len = GENERATE_LENGTH,
            cache_kv = True
        )

        output_str = decode_tokens(sample)
        print(f'{output_str}\n\n')



import tqdm
import torch
import torch.optim as optim
import torch.nn.functional as F

from x_transformers import TransformerWrapper, Decoder

# constants

BATCH_SIZE = 256
LEARNING_RATE = 3e-4
EVAL_EVERY  = 500

EVAL_LENGTHS = (16, 32, 64, 128, 256, 512)
TRAIN_MAX_LENGTH = EVAL_LENGTHS[-2]

LOSS_THRES_INCREASE_LEN = 1e-3
MEET_CRITERIA_THRES_INCREASE_LEN = 10

HYBRIDIZE_WITH_RNN = True

# rnn for fully resolving state tracking by hybridization
# but will also look into gated delta net + negative eigenvalues (Songlin Yang et al) as a parallel solution

dim = 64
heads = 4
dim_head = 32
decoder_kwargs = dict()

if HYBRIDIZE_WITH_RNN:
    from torch.nn import GRU

    decoder_kwargs = dict(
        attn_hybrid_fold_axial_dim = 4, # even if recurrence is every 4 tokens, can generalize for parity
        attn_hybrid_learned_mix = True,
        attn_hybrid_module = GRU(dim, dim_head * heads, batch_first = True)
    )

# instantiate model

model = TransformerWrapper(
    num_tokens = 2,
    max_seq_len = 0,
    attn_layers = Decoder(
        dim = dim,
        depth = 3,
        heads = heads,
        attn_dim_head = dim_head,
        shift_tokens = 1, # helps a lot with parity training, but not able to generalize on its own
        **decoder_kwargs
    )
).cuda()

# optimizer

from lion_pytorch.cautious_lion import Lion

optimizer = Lion(model.parameters(), lr = LEARNING_RATE, cautious_factor = 0.1)

# data generator

def cycle(length):
    while True:
        seq = torch.randint(0, 2, (BATCH_SIZE, length)).cuda()
        labels = (seq.cumsum(dim = -1) % 2)
        yield (seq, labels)

# dataloaders

train_dl = cycle(TRAIN_MAX_LENGTH)

eval_dls = {eval_length: cycle(eval_length) for eval_length in EVAL_LENGTHS}

print(f'training at max length: {TRAIN_MAX_LENGTH}')

# training

i = 0
meet_criteria = 0
train_seq_len = 1
stop_length = EVAL_LENGTHS[-2]

with tqdm.tqdm(mininterval = 10., desc = 'training') as pbar:

    while train_seq_len < stop_length:
        model.train()

        seq, labels = next(train_dl)

        # length curriculum learning

        seq = seq[:, :train_seq_len]
        labels = labels[:, :train_seq_len]

        logits = model(seq)

        loss = F.cross_entropy(logits.transpose(-1, -2), labels, reduction = 'none')
        last_loss = loss[:, -1].mean()
        loss.mean().backward()

        if last_loss.item() < LOSS_THRES_INCREASE_LEN:
            meet_criteria += 1
        else:
            meet_criteria = 0

        if meet_criteria >= MEET_CRITERIA_THRES_INCREASE_LEN:
            meet_criteria = 0
            train_seq_len += 1
            print(f'criteria met, incrementing to {train_seq_len}')

        print(f'({train_seq_len})| {i}: {last_loss.item()}')
        torch.nn.utils.clip_grad_norm_(model.parameters(), 0.5)

        optimizer.step()
        optimizer.zero_grad()

        last_step = train_seq_len == stop_length

        if last_step:
            print(f'made it to training length {train_seq_len}. running final eval to check for generalization')

        if last_step or (i + 1) % EVAL_EVERY == 0:

            model.eval()
            print('\n')

            for eval_length, eval_dl in eval_dls.items():
                incorrects = 0

                seq, labels = next(eval_dl)

                logits = model(seq)
                pred = logits[:, -1].argmax(dim = -1)
                incorrects = (pred != labels[:, -1]).abs().sum().item()

                frac_incorrect = incorrects * 100 / BATCH_SIZE

                print(f"{eval_length}\t - frac incorrect:\t {frac_incorrect:.1f}%")

            print('\n')

        i += 1
        pbar.update(1)




# Belief State Transformer

# Hu et al. https://arxiv.org/abs/2410.23506
# https://www.youtube.com/watch?v=aqhbRtB2Fyg

from __future__ import annotations

import torch
from torch.autograd import Function
from torch.nn import Module, ModuleList
from torch import nn, cat, stack, tensor, arange, cartesian_prod
import torch.nn.functional as F

from x_transformers.autoregressive_wrapper import (
    eval_decorator,
    min_p,
)

from x_transformers.x_transformers import (
    Decoder,
    TransformerWrapper
)

import einx
from einops import rearrange, repeat, pack, unpack
from einops.layers.torch import Rearrange

# helper functions

def exists(v):
    return v is not None

def default(v, d):
    return v if exists(v) else d

# wrappers

class BeliefStateWrapper(Module):
    """
    Figure 13. in https://arxiv.org/abs/2410.23506
    """

    def __init__(
        self,
        forward_decoder: TransformerWrapper,
        backward_decoder: TransformerWrapper | None = None,
        train_frac_forward_backward_pairs: float = 1.,
        text_head: Module | None = None,
        backward_ar_loss_weight: float = 1., # can weigh the training of the backwards decoder differently, perhaps fwd/bwd have a shared backbone etc etc
        pred_terminal = False,
        pred_terminal_loss_weight: float = 1.
    ):
        super().__init__()
        backward_decoder = default(backward_decoder, forward_decoder) # if backward decoder not set, use the same transformer, assume it knows how to switch gears based on suffix token

        assert forward_decoder.emb_dim == backward_decoder.emb_dim, 'forward and backwards model must have the same embedding dimension'
        assert forward_decoder.num_tokens == backward_decoder.num_tokens, 'forward and backwards model must have the same number of tokens'

        dim = forward_decoder.emb_dim
        num_tokens = forward_decoder.num_tokens

        self.num_tokens = num_tokens

        # the suffix token

        self.suffix_token = nn.Parameter(torch.zeros(dim))
        nn.init.normal_(self.suffix_token, std = 0.02)

        # the text prediction head, which predicts for the combinations of prefix and suffix the next and previous token for forwards and backward sequences

        if not exists(text_head):
            text_head = nn.Sequential(
                nn.Linear(dim * 2, dim),
                nn.LeakyReLU(),
                nn.Linear(dim, num_tokens * 2),
            )

        self.text_head = text_head

        # predicting terminal state (when suffix and prefix predict the same token)

        self.to_terminal_logit = nn.Sequential(
            nn.Linear(dim * 2, dim),
            nn.LeakyReLU(),
            nn.Linear(dim, 1),
            Rearrange('... 1 -> ...')
        ) if pred_terminal else None

        self.pred_terminal_loss_weight = pred_terminal_loss_weight

        # the two decoders, one which is causal forward, the other causal backwards

        self.forward_decoder = forward_decoder
        self.backward_decoder = backward_decoder

        # what fraction of forward backward pairs to train on
        # for further memory efficiency

        assert 0 < train_frac_forward_backward_pairs <= 1.
        self.train_frac_fb_pairs = train_frac_forward_backward_pairs
        self.needs_subsample_fb_pairs = train_frac_forward_backward_pairs < 1.

        # loss weighting

        self.backward_ar_loss_weight = backward_ar_loss_weight
        self.needs_loss_weight = backward_ar_loss_weight != 1.

        self.register_buffer('loss_weights', tensor([1., self.backward_ar_loss_weight]))

        # sampling

        self.max_seq_len = self.forward_decoder.max_seq_len

    @torch.no_grad()
    @eval_decorator
    def generate_with_suffix_cond(
        self,
        prompts,
        seq_len,
        temperature = 1.25,
        cache_kv = False,
        suffix: Tensor | None = None, # the goal conditioning
        filter_logits_fn = min_p,
        filter_kwargs = dict(
            min_p = 0.1
        ),
        decode_backwards = False,
        **kwargs
    ):
        max_seq_len, greedy, device = self.max_seq_len, temperature == 0., prompts.device

        prompts, batch_ps = pack([prompts], '* d')

        batch, orig_seq_len = prompts.shape

        # allow for decoding backwards, to make sure it is working

        main_decoder = self.forward_decoder

        if decode_backwards:
            prompts = prompts.flip(1)
            main_decoder = self.backward_decoder

        out = prompts

        # kv caches

        cache = None

        # get the encoded suffix token once

        suffix_sos_tokens = rearrange(self.suffix_token, 'd -> 1 1 d')

        suffix_sos_tokens = repeat(suffix_sos_tokens, '1 1 d -> b 1 d', b = batch)

        if not decode_backwards:
            if exists(suffix):
                if suffix.ndim == 1:
                    suffix = repeat(suffix, 'n -> b n', b = batch)

                suffix = suffix.flip(1) # reverse autoregressive

            suffix_embed = self.backward_decoder(
                suffix,
                prepend_embeds = suffix_sos_tokens,
                return_embeddings = True
            )

            # pick out the last embedding for fill in the middle

            suffix_embed = suffix_embed[:, -1:]

        else:
            # just grab a random token for now for prefix

            prefix_embed = torch.randint(0, self.num_tokens, (batch, 1), device = device)

            prefix_embed = self.forward_decoder(prefix_embed, return_embeddings = True)

        # sampling up to seq_len

        for _ in range(seq_len):

            embeds, new_cache = main_decoder(
                out,
                prepend_embeds = suffix_sos_tokens if decode_backwards else None,
                return_intermediates = True,
                return_embeddings = True,
                cache = cache,
                **kwargs
            )

            last_embeds = embeds[:, -1:]

            if not decode_backwards:
                embeds = cat((last_embeds, suffix_embed), dim = -1)
            else:
                embeds = cat((prefix_embed, last_embeds), dim = -1)

            if cache_kv and self.forward_decoder.can_cache_kv:
                cache = new_cache

            forward_logits, backward_logits = self.text_head(embeds).chunk(2, dim = -1)

            logits = forward_logits if not decode_backwards else backward_logits

            logits = logits[:, -1]

            if greedy:
                sample = logits.argmax(dim = -1, keepdim = True)
            else:
                filtered_logits = filter_logits_fn(logits, **filter_kwargs)
                probs = F.softmax(filtered_logits / temperature, dim = -1)
                sample = torch.multinomial(probs, 1)

            # concat sample

            out = torch.cat((out, sample), dim = -1)

        out = out[:, orig_seq_len:]

        out, = unpack(out, batch_ps, '* n')

        return out

    def forward(
        self,
        seq,
        return_loss_only = False,
        loss_scale = 1.
    ):
        batch, seq_len, device = *seq.shape, seq.device

        # forward autoregressive

        forward_embeds = self.forward_decoder(seq, return_embeddings = True)

        # backward autoregressive

        backward_seq = seq.flip(1)

        suffix_tokens = repeat(self.suffix_token, 'd -> b 1 d', b = batch)

        backward_embeds = self.backward_decoder(
            backward_seq,
            prepend_embeds = suffix_tokens,
            return_embeddings = True
        )

        backward_embeds = backward_embeds.flip(1)

        # trick to reduce memory on backwards pass

        orig_forward_embeds, forward_embeds = forward_embeds, forward_embeds.detach()
        orig_backward_embeds, backward_embeds = backward_embeds, backward_embeds.detach()

        forward_embeds.requires_grad_()
        backward_embeds.requires_grad_()

        # belief state objective

        seq_arange = arange(seq_len, device = device)

        fb_pairs = cartesian_prod(seq_arange, seq_arange + 1) # plus one for suffix token

        # filter down to valid pairs, as in figure 11
        # f - forward, b - backward, i - indices

        fi, bi = fb_pairs.unbind(dim = -1)

        valid_mask = (bi - fi) >= 2

        fb_pairs = fb_pairs[valid_mask]

        # maybe subsample fb pairs

        if self.needs_subsample_fb_pairs:
            num_pairs = fb_pairs.shape[0]

            num_subsampled = max(int(num_pairs * self.train_frac_fb_pairs), 1)

            rand_subsampled_indices = torch.randperm(num_pairs, device = device)[:num_subsampled]

            fb_pairs = fb_pairs[rand_subsampled_indices]

        # get labels for both

        fi, bi = fb_pairs.unbind(dim = -1)

        labels_fi, labels_bi = (fi + 1), (bi - 1)

        forward_labels, backward_labels = seq[:, labels_fi], seq[:, labels_bi]

        labels = cat((forward_labels, backward_labels), dim = -1)

        # get the forward and backward embedding pairs and feed them through the text head for both forward and backward predictions

        fb_embeds = cat((
            forward_embeds[:, fi],
            backward_embeds[:, bi]
        ), dim = -1)

        logits = self.text_head(fb_embeds)

        # cross entropy loss

        loss = F.cross_entropy(
            rearrange(logits, 'b n (fb l) -> b l (fb n)', fb = 2),
            labels,
            reduction = 'none' if self.needs_loss_weight else 'mean'
        )

        # maybe predict terminal

        if exists(self.to_terminal_logit):
            terminal_logits = self.to_terminal_logit(fb_embeds)

            terminal_labels = ((bi - fi) == 2).float() # distance is exactly 2
            terminal_labels = repeat(terminal_labels, 'n -> b n', b = batch)

            is_end_loss = F.binary_cross_entropy_with_logits(
                terminal_logits,
                terminal_labels
            )

            loss = (
                loss +
                is_end_loss * self.pred_terminal_loss_weight
            )

        # maybe early return loss

        if return_loss_only:
            return loss

        # maybe loss weighting

        if self.needs_loss_weight:
            loss = rearrange(loss, 'b (fb n) -> b fb n', fb = 2)
            loss = einx.multiply('b fb n, fb', loss, self.loss_weights)
            loss = loss.mean()

        # backwards

        (loss * loss_scale).backward()

        orig_forward_embeds.backward(forward_embeds.grad)
        orig_backward_embeds.backward(backward_embeds.grad)

        return loss



import tqdm
import torch
import torch.optim as optim
from x_transformers import XTransformer

# constants

NUM_BATCHES = int(1e5)
BATCH_SIZE = 32
LEARNING_RATE = 3e-4
GENERATE_EVERY  = 100
NUM_TOKENS = 16 + 2
ENC_SEQ_LEN = 32
DEC_SEQ_LEN = 64 + 1

# helpers

def cycle():
    while True:
        prefix = torch.ones((BATCH_SIZE, 1)).long().cuda()
        src = torch.randint(2, NUM_TOKENS, (BATCH_SIZE, ENC_SEQ_LEN)).long().cuda()
        tgt = torch.cat((prefix, src, src), 1)
        src_mask = torch.ones(BATCH_SIZE, src.shape[1]).bool().cuda()
        yield (src, tgt, src_mask)

# instantiate model

model = XTransformer(
    dim = 512,
    tie_token_emb = True,
    return_tgt_loss = True,
    enc_num_tokens=NUM_TOKENS,
    enc_depth = 3,
    enc_heads = 8,
    enc_max_seq_len = ENC_SEQ_LEN,
    dec_num_tokens = NUM_TOKENS,
    dec_depth = 3,
    dec_heads = 8,
    dec_max_seq_len = DEC_SEQ_LEN
).cuda()

# optimizer

optim = torch.optim.Adam(model.parameters(), lr=LEARNING_RATE)

# training

for i in tqdm.tqdm(range(NUM_BATCHES), mininterval=10., desc='training'):
    model.train()

    src, tgt, src_mask = next(cycle())

    loss = model(src, tgt, mask=src_mask)
    loss.backward()
    print(f'{i}: {loss.item()}')

    optim.step()
    optim.zero_grad()

    if i != 0 and i % GENERATE_EVERY == 0:
        model.eval()
        src, _, src_mask = next(cycle())
        src, src_mask = src[:1], src_mask[:1]
        start_tokens = (torch.ones((1, 1)) * 1).long().cuda()

        sample = model.generate(src, start_tokens, ENC_SEQ_LEN, mask = src_mask)
        incorrects = (src != sample).abs().sum()

        print(f"input:  ", src)
        print(f"predicted output:  ", sample)
        print(f"incorrects: {incorrects}")
